from mext.threads.locks import SimpleLock
from mext.threads.sync import Event

from collections import deque
from time import time
import sys, heapq

__all__ = ['Queue', 'PriorityQueue', 'LifoQueue', 'Empty']


class Empty(Exception):
    "Exception raised by Queue.get(block=0)."
    pass


class Queue(object):
    def __init__(self, maxsize=0):
        self.maxsize = maxsize or sys.maxint
        self._init(maxsize)
        self.lock = SimpleLock()
        # threads waiting to get an item wait on this condition
        self._evt_has_items = Event(self.lock)
        # threads waiting to put an item wait on this one
        self._evt_has_space = Event(self.lock)
        # threads waiting on wait_tasks() are waiting on this one
        self._evt_tasks = Event(self.lock)
        self.tasks = 0 # tasks in queue or being processed

    def task_done(self):
        with self.lock:
            if not self.tasks:
                raise ValueError('task_done() called too many times')
            self.tasks -= 1
            if not self.tasks:
                self._evt_tasks.notify(None)

    def wait_tasks(self):
        with self.lock:
            while self.tasks:
                self._evt_tasks.wait()
    join = wait_tasks

    def __len__(self):
        return len(self.queue)

    @property
    def is_full(self):
        return len(self.queue) >= self.maxsize

    def put(self, item, timeout='never'):
        with self.lock:
            if not _wait_timeout(
                self._evt_has_space,
                lambda: len(self.queue) < self.maxsize,
                timeout
            ):
                return False
            self._put(item)
            self.tasks += 1
            self._evt_has_items.notify(1)
            return True

    def get(self, timeout='never'):
        with self.lock:
            if not _wait_timeout(
                self._evt_has_items,
                lambda: len(self.queue),
                timeout
            ):
                raise Empty
            self._evt_has_space.notify(1)
            return self._get()

    def _init(self, maxsize):
        q = self.queue = deque()
        self._put = q.append
        self._get = q.popleft


class PriorityQueue(Queue):
    def _init(self, maxsize):
        self.queue = []

    def _put(self, item, heappush=heapq.heappush):
        heappush(self.queue, item)

    def _get(self, heappop=heapq.heappop):
        return heappop(self.queue)


class LifoQueue(Queue):
    def _init(self, maxsize):
        q = self.queue = []
        self._put = q.append
        self._get = q.pop



def _wait_timeout(cond, check, timeout):
    if timeout == 'never': # wait for as long as necessary
        while not check():
            cond.wait()
        return True
    elif not timeout: # nonblocking check
        return check()
    elif timeout < 0:
        raise ValueError("'timeout' must be a positive number")
    else:
        endtime = time() + timeout
        while not check():
            remaining = endtime - time()
            if remaining <= 0:
                return False
            cond.wait(remaining)
        return True

